import StringIO
import hashlib
import json
import logging
import zipfile

from pygments import highlight, lexers, formatters

logger = logging.getLogger('koodous-api')

try:
    from androguard.core.bytecodes.apk import APK
except ImportError:
    APK = None


def sha256(filepath):
    """
        Function to calculate sha256 of a file in disk
        Params:
            - filepath (str): Path where the file is stored
        Return:
            - sha256 hash calculated
    """
    fd = open(filepath, 'rb')
    hasher_256 = hashlib.sha256()
    chunk = fd.read(1024)
    while chunk:
        hasher_256.update(chunk)
        chunk = fd.read(1024)

    fd.close()
    return hasher_256.hexdigest()


def unpack(filepath, dst):
    """
        Function to generate a file from uncompressed file of an APK.
        Params:
            - filepath (str): Path where the APK is stored
            - dst (str): Path where the file generated will be saved.
        Return:
            - True is all was done.
            - False in case of error.
    """

    to_ret = True

    with open(filepath, 'rb') as fd:
        content = fd.read()

    try:
        zipped = zipfile.ZipFile(filename)
        for name in zipped.namelist():
            content += ' ' + zipped.read(name)

    except:
        to_ret = False

    with open(dst, 'wb') as fd:
        fd.write(content)

    return to_ret


def is_apk_file(filepath):
    """
        Function to check if a file is an APK
        Params:
            - filepath (str): Path where is the file to check
        Return:
            True (bool) if the file is an APK
            False (bool) otherwise
    """
    try:
        zip = zipfile.ZipFile(open(filepath))
        for i in zip.namelist():
            if i == 'AndroidManifest.xml':
                return True

        return False
    except:
        return False


def is_apk(content):
    """
        Function to check if a content is an APK
        Params:
            - content (str): Binary data of the file to check
        Return:
            True (bool) if the file is an APK
            False (bool) otherwise
    """
    try:
        zip = zipfile.ZipFile(StringIO.StringIO(content))
        for i in zip.namelist():
            if i == 'AndroidManifest.xml':
                return True

        return False
    except:
        return False


def pygmentize(output, lexer=lexers.JsonLexer()):
    """
    Pygmentize any string for pretty terminal printing.

    :param lexer: a lexer from `pygments.lexers`
    :param output: the string to print
    :return: the pygmentized string
    """

    return highlight(unicode(output, 'UTF-8'), lexer,
                     formatters.TerminalFormatter())


def pygmentize_json(obj, sort_keys=True, indent=4):
    """
    Pretty print a JSON string from a dictionary

    :param indent: indentation spaces
    :param sort_keys: whether to sort the keys
    :param obj: a dict to build the JSON output string from
    :return: a pygmentized string
    """
    return pygmentize(json.dumps(obj, sort_keys=sort_keys, indent=indent))


def is_apk(filepath):
    """
    Check whether filepath is pointing to a readable, valid APK file.

    Requires androguard. If androguard is not installed, fallsback to
    returning always `True`.

    :param filepath: path to file
    :return: `True` if the path is pointing to a readable, valid APK file
    """
    if not APK:
        return True

    try:
        APK(filepath)
        return True
    except Exception as ex:
        logger.warning('File %s does not seem a valid APK file', filepath)

    return False


#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2015. The Koodous Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import hashlib
import logging
import time
import urllib

import certifi
import requests

import utils

__author__ = "Antonio Sanchez <asanchez@koodous.com>"

BASE_URL = 'https://koodous.com/api/'

REQUESTS_CA_BUNDLE = certifi.where()

logger = logging.getLogger('koodous-api')


class Koodous(object):
    def __init__(self, token):
        self.token = token
        self.headers = {'Authorization': 'Token %s' % token}

    def my_user(self):
        url = '%s/analysts/current' % BASE_URL
        response = requests.get(url=url, headers=self.headers,
                                verify=REQUESTS_CA_BUNDLE)
        if response.status_code == 200:
            return response.json()

        return None

    def upload(self, filepath):
        """
            Function to upload a file
        """

        sha256_file = utils.sha256(filepath)
        url = '%s/apks/%s/get_upload_url' % (BASE_URL, sha256_file)
        response = requests.get(url=url, headers=self.headers,
                                verify=REQUESTS_CA_BUNDLE)
        if response.status_code == 200:
            json_data = response.json()
            # print json_data.get('upload_url', None)
            files = {'file': open(filepath, 'rb')}

            response = requests.post(url=json_data.get("upload_url"),
                                     files=files,
                                     verify=REQUESTS_CA_BUNDLE)
            while response.status_code == 404:  # Workaround server problem sometimes
                time.sleep(1)
                response = requests.post(url=json_data.get("upload_url"),
                                         files=files,
                                         verify=REQUESTS_CA_BUNDLE)
            return sha256_file
        elif response.status_code == 409:
            raise Exception("APK already exists")
        else:
            raise Exception("Unknown error: %s" % response.text)

    def download_to_file(self, sha256, dst):
        """
            Function to download a file to a destination
            Params:
                - sha256 (str): sha256 hash required.
                - dst (str): Path where the sample will be saved.
            Return:
                - sha256 hash if all was done.
            Exception:
                - Exception() with text "Something was wrong during download"
        """
        down_url = self.get_download_url(sha256)
        if down_url:
            res = requests.get(url=down_url, verify=REQUESTS_CA_BUNDLE)
            sha256_downloaded = hashlib.sha256(res.content).hexdigest()

            if sha256 != sha256_downloaded:
                raise Exception("Something was wrong during download")

            with open(dst, "wb") as fd:
                fd.write(res.content)

        else:
            raise Exception("Something was wrong during download")
        return sha256

    def get_download_url(self, sha256):
        """
            Return download URL for hash "sha256"
            Params:
                - sha256 (str): sha256 hash required.
            Return:
                - str: Download URL (valid during 5 minutes).
        """
        url = '%s/apks/%s/download' % (BASE_URL, sha256)

        response = requests.get(url=url, headers=self.headers,
                                verify=REQUESTS_CA_BUNDLE)
        if response.status_code == 200 and response.json().get('download_url',
                                                               None):
            return response.json().get('download_url', None)
        else:
            return None

    def search(self, search_term, limit=None):
        """
            Return apks objects (in json) list according a search string
            Params:
                - search_term (str): Search string according Koodous documentation
                - limit (int): Number to return only the first "limit" apks
            Return:
                - list: apks information found by the search.
        """
        to_ret = list()

        if not limit:
            limit = float('Inf')

        next_page = '%s/apks?page_size=100&search=%s' % (BASE_URL,
                                                         urllib.quote(
                                                             search_term))

        while next_page and len(to_ret) < limit:
            response = requests.get(url=next_page,
                                    headers=self.headers,
                                    verify=REQUESTS_CA_BUNDLE)
            next_page = response.json().get('next', None)
            for sample in response.json().get('results', []):
                to_ret.append(sample)
                if len(to_ret) == limit:
                    break
        return to_ret

    def get_analysis(self, sha256):
        url = '%s/apks/%s/analysis' % (BASE_URL, sha256)

        response = requests.get(url=url, headers=self.headers,
                                verify=REQUESTS_CA_BUNDLE)
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 405: #Analysis doesn't exist
            raise Exception(
                  "This sample has not analysis available, you can request it.")
        # Otherwise
        return None

    def get_public_ruleset(self, ruleset_id):
        """
        Retrieve a public ruleset by id.

        :param ruleset_id: identifier of the public ruleset.
        :return: `dict`
        """
        url = '{endpoint}/public_rulesets/{id}'.format(**dict(
            endpoint=BASE_URL,
            id=ruleset_id))
        response = requests.get(url=url, headers=self.headers,
                                verify=REQUESTS_CA_BUNDLE)
        if response.status_code == 200:
            return response.json()
        return None

    def iter_matches_public_ruleset(self, ruleset_id):
        """
        Generator over the the matches of a public ruleset by id.

        :param ruleset_id: identifier of the public ruleset.
        :return: generator
        """
        next_page = '{endpoint}/ruleset_matches/{id}/apks'.format(**dict(
            endpoint=BASE_URL,
            id=ruleset_id))

        while next_page:
            response = requests.get(url=next_page, headers=self.headers,
                                    verify=REQUESTS_CA_BUNDLE)
            if response.status_code != 200:
                break

            r = response.json().get("results", None)

            if r:
                yield r

            next_page = response.json().get('next', None)

    def get_matches_public_ruleset(self, ruleset_id):
        """
        Retrieve the matches of a public ruleset by id. Avoid using this
        method on rulesets with many matches. Use the iterator version
        instead.

        :param ruleset_id: identifier of the public ruleset.
        :return: `list`
        """
        results = []

        for r in self.iter_matches_public_ruleset(ruleset_id):
            results.extend(r)

        return results

    def analyze(self, sha256):
        url = '%s/apks/%s/analyze' % (BASE_URL, sha256)
        res = requests.get(url=url, headers=self.headers,
                           verify=REQUESTS_CA_BUNDLE)
        if res.status_code == 200:
            return True
        return False

    def post_comment(self, sha256, comment):
        url = '%s/apks/%s/comments' % (BASE_URL, sha256)
        payload = {'text': comment}

        response = requests.post(url=url, headers=self.headers,
                                 data=payload, verify=REQUESTS_CA_BUNDLE)

        if response.status_code == 201:
            return response.json().get('text', None)

        return None

    def get_comments(self, sha256):
        to_ret = list()
        next_page = '%s/apks/%s/comments' % (BASE_URL, sha256)

        while next_page:
            response = requests.get(url=next_page,
                                    headers=self.headers,
                                    verify=REQUESTS_CA_BUNDLE)
            next_page = response.json().get('next', None)
            for sample in response.json().get('results', []):
                to_ret.append(sample)

        return to_ret

    def delete_comment(self, comment_id):
        url = '%s/comments/%s' % (BASE_URL, comment_id)
        response = requests.delete(url=url, headers=self.headers,
                                   verify=REQUESTS_CA_BUNDLE)

        if response.status_code == 204:
            return True

        return False

    def vote_apk_positive(self, sha256, kind):
        """
            To send a positive vote to an APK (goodware)
        """
        url = '%s/apks/%s/votes' % (BASE_URL, sha256)
        return requests.post(url, data={'kind': 'positive'},
                             headers=self.headers, verify=REQUESTS_CA_BUNDLE)

    def vote_apk_negative(self, sha256, kind):
        """
            To send a negative vote to an APK (malware)
        """

        url = '%s/apks/%s/votes' % (BASE_URL, sha256)
        return requests.post(url, data={'kind': 'negative'},
                             headers=self.headers, verify=REQUESTS_CA_BUNDLE)

    def vote_apk(self, sha256, kind):
        """
            Function to send a positive or negative vote to an APK.
            Params:
                sha256 (str): sha256 of the file
                kind (str): 'positive' to send a positive vote or
                            'negative' to send a negative vote
            Return:
                TODO
        """
        kind = str(kind)
        if kind != 'positive' or kind != negative:
            raise Exception("Kind vote must be positive or negative")

        url = '%s/apks/%s/votes' % (BASE_URL, sha256)

        return requests.post(url, data={'kind': kind},
                             headers=self.headers, verify=REQUESTS_CA_BUNDLE)



#!/usr/bin/env python
import pip
from pip.req import parse_requirements
from setuptools import setup

REQUIREMENTS_FILE = 'requirements.txt'

try:
    install_reqs = parse_requirements(REQUIREMENTS_FILE,
                                      session=pip.download.PipSession())
except AttributeError as ex:
    raise ex

    import warnings
    warnings.warn('Your PIP is very old: we strongly recommend to update it')

reqs = [str(ir.req) for ir in install_reqs]

setup(name='koodous-py',
      packages=['koodous'],
      version='0.4',
      description='Module to interact with Koodous API',
      author='Antonio Sanchez',
      author_email='asanchez@koodous.com',
      license='Apache Version 2',
      url='https://koodous.com/',
      download_url='https://github.com/Koodous/python-sdk/archive/master.zip',
      keywords=['koodous', 'api', 'sdk', 'python', 'android', 'apk', 'malware'],
      install_requires=reqs,
      entry_points='''
        [console_scripts]
        koocli=koodous.cli:cli
      '''
      )



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2015. The Koodous Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import glob
import io
import json
import logging
import os

import click

from koodous import Koodous
from utils import pygmentize_json, is_apk, sha256 as file_hash

__author__ = "Federico Maggi <federico.maggi@gmail.com>"

logger = logging.getLogger('koocli')

LOGGING_LEVELS = dict((
    ('notset', logging.NOTSET),
    ('debug', logging.DEBUG),
    ('info', logging.INFO),
    ('warning', logging.WARNING),
    ('error', logging.ERROR),
    ('critical', logging.CRITICAL),
))

loglevels = LOGGING_LEVELS.keys()


@click.group()
@click.option('--quiet/--no-quiet', default=False,
              help='Suppress output (logging is configured separately)')
@click.option('--wdir', type=click.Path(exists=False), required=True,
              help='Working directory')
@click.option('--loglevel', type=click.Choice(loglevels), default='info')
@click.option('--token', envvar='TOKEN', help='Koodous API token',
              required=True)
def cli(quiet, loglevel, wdir, token):
    """A simple command line interface (CLI) to the Koodous API.

    In order to use this CLI, you need an account at koodous.com and you
    need to grab your API token at https://koodous.com/settings/profile

    You can pass the API token both as a command line option, or set it as
    an environment variable (TOKEN).

    \b To get help for each individual command, just type

        $ koocli <command_name> --help

    """
    ctx = click.get_current_context()

    ctx.meta['TOKEN'] = token
    ctx.meta['api'] = Koodous(token=token)
    ctx.meta['wdir'] = wdir
    ctx.meta['quiet'] = quiet

    import coloredlogs
    coloredlogs.install(level=LOGGING_LEVELS[loglevel])


@cli.command()
@click.option('--save/--no-save', default=False,
              help='Save JSON to file in the working directory')
@click.option('--outfile', help='Optional file to save the metadata to',
              type=click.File('wb'))
@click.argument('ruleset_id', type=int)
def get_public_ruleset(save, outfile, ruleset_id):
    """Get a public ruleset by its RULESET_ID"""
    ctx = click.get_current_context()

    api = ctx.meta.get('api')
    wdir = ctx.meta.get('wdir')
    quiet = ctx.meta.get('quiet')

    logger.info('Attempting to fetch ruleset %s', ruleset_id)
    result = api.get_public_ruleset(ruleset_id=ruleset_id)

    if not quiet:
        click.echo(pygmentize_json(result))

    if save:
        if not outfile:
            filepath = os.path.join(wdir, 'ruleset-{}.json'.format(ruleset_id))
            outfile = io.open(filepath, 'wb')
        else:
            filepath = outfile.name
        logger.info('Saving ruleset metadata to {}'.format(filepath))

        json.dump(result, outfile)

        logger.info('Ruleset metadata saved correctly')


@cli.command()
@click.option('--save/--no-save', default=False,
              help='Save output as separate JSON files in the working '
                   'directory')
@click.option('--download/--no-download', default=False,
              help='Download the actual APK files and save them in the '
                   'working directory')
@click.option('--prompt/--no-prompt', default=True,
              help='Prompt for confirmations')
@click.option('--limit', type=int, default=0,
              help='Stop after LIMIT matches (defaults to 0, no limits)')
@click.argument('ruleset_id', type=int)
def get_matches_public_ruleset(ruleset_id, prompt, save, download, limit):
    """Get the APKs that match a public ruleset by its RULESET_ID

    Example: https://koodous.com/rulesets/RULESET_ID (e.g., 666)
    """
    ctx = click.get_current_context()

    api = ctx.meta.get('api')
    quiet = ctx.meta.get('quiet')
    wdir = ctx.meta.get('wdir')

    logger.info('Attempting to fetch ruleset %s', ruleset_id)
    ruleset = api.get_public_ruleset(ruleset_id=ruleset_id)

    d = ruleset['detections']

    if save:
        filepath = os.path.join(wdir, 'ruleset-{}.json'.format(
            ruleset_id))

        logger.info('Saving ruleset to %s', filepath)
        with io.open(filepath, 'wb') as outfile:
            json.dump(ruleset, outfile)

        logger.info('Ruleset saved successfully')

    if prompt and 100 < d <= limit:
        if not click.confirm('The selected ruleset has {} matches. Do you '
                             'want to proceed printing all of '
                             'them?'.format(d)):
            return

    iterator = api.iter_matches_public_ruleset(ruleset_id)
    count = 0

    for apks in iterator:
        for apk in apks:
            if not quiet:
                click.echo(pygmentize_json(apk))

            if save:
                sha256 = apk['sha256']

                filepath = os.path.join(wdir, '{}.json'.format(sha256))

                logger.info('Saving metadata of %s to %s', sha256, filepath)

                with io.open(filepath, 'wb') as outfile:
                    json.dump(apk, outfile)

            if download:
                dst = os.path.join(wdir, '{}.apk'.format(sha256))
                logger.info('Downloading %s to %s', sha256, dst)

                try:
                    api.download_to_file(sha256=sha256, dst=dst)
                    logger.info('APK downloaded successfully')
                except Exception as ex:
                    logger.error('Could not download %s: %s', sha256, ex)

            count += 1

            if 0 < limit <= count:
                logger.info('Limit of %s matches reached: stopping!', limit)
                return


@cli.command()
@click.argument('glob_exp')
@click.option('--ignore-bad-apks/--no-ignore-bad-apks',
              default=False, help='Bypass APK checks')
def upload(glob_exp, ignore_bad_apks):
    """Upload files matching the GLOB_EXP to Koodous.

    \b
    Example of GLOB_EXP are:
    \b
        - /path/to/*-files/*.apk
        - /path/*

    Some shell interpreters require you to wrap glob expression into
    single quotes to prevent automatic expansion:

        $ TOKEN='<your token>' koocli --wdir /tmp/ upload '/path/*.apk'

    Note that non-files will be obviously skipped. And we do try
    to skip non-APK files. Sometimes, however, malformed APKs exists that
    manage to bypass or fail the validation.
    """
    ctx = click.get_current_context()
    api = ctx.meta.get('api')
    quiet = ctx.meta.get('quiet')

    count = 0
    skipped = 0
    errors = 0

    for path in glob.glob(glob_exp):
        if os.path.isfile(path) and os.access(path, os.R_OK):
            if is_apk(path) or ignore_bad_apks:
                logger.info('Submitting file %s to Koodous', path)

                sha256 = None
                try:
                    sha256 = api.upload(path)
                except Exception as ex:
                    logger.error('Error uploading file %s: %s', path, ex)
                    errors += 1

                if sha256:
                    logger.info('File %s (%s) uploaded correctly', path,
                                sha256)

                    if not quiet:
                        click.echo(sha256)

                    count += 1
            else:
                logger.warning('Skipping file %s: invalid APK format', path)
                skipped += 1

    logger.info('Correctly submitted %d files (skipped %d, errors %d)',
                count, skipped, errors)


@cli.command()
@click.option('--upload/--no-upload', default=False,
              help='Upload the file for analysis if not found')
@click.option('--save/--no-save', default=False,
              help='Save JSON to file in the working directory')
@click.option('--outfile', help='Optional file to save the metadata to',
              type=click.File('wb'))
@click.argument('sha256_or_file')
def get_analysis(sha256_or_file, upload, save, outfile):
    """
    Get the Koodous report of SHA256_OR_FILE. If the file has not be analyzed
    by Koodous, the file is just submitted (or not, according to the
    --upload option).
    """
    ctx = click.get_current_context()
    api = ctx.meta.get('api')
    wdir = ctx.meta.get('wdir')

    is_file = os.path.isfile(sha256_or_file) and os.access(sha256_or_file,
                                                           os.R_OK)
    sha256 = sha256_or_file
    if is_file:
        sha256 = file_hash(sha256_or_file)
        logger.info('File %s SHA-256 digest = %s', sha256_or_file, sha256)

    logger.info('Getting analysis of %s', sha256)

    analysis = api.get_analysis(sha256)

    click.echo(analysis)

    if analysis:
        click.echo(pygmentize_json(analysis))

        if save:
            if not outfile:
                filepath = os.path.join(wdir, '{}.json'.format(sha256))
                outfile = io.open(filepath, 'wb')
            else:
                filepath = outfile.name

            logger.info('Saving analysis to %s', filepath)
            json.dump(analysis, outfile)

            logger.info('Saved to %s successfully', filepath)
    elif is_file:
        logger.warning('File not found on Koodous')

        if upload:
            logger.info('Uploading file for analysis')

            try:
                upload_result = api.upload(sha256_or_file)
                logger.info('File %s uploaded successfully', upload_result)
            except Exception as ex:
                logger.error('Uploading %s failed: %s', sha256_or_file, ex)
